from .projection import project
from shapely.geometry import (
    GeometryCollection,
    LinearRing,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)
from shapely.ops import cascaded_union


kind_mapping = {
    'Polygon': 'polygon',
    'MultiPolygon': 'polygon',
    'LineString': 'line',
    'MultiLineString': 'line',
    'LinearRing': 'line',
    'Point': 'point',
    'MultiPoint': 'point',
}


class IncompatibleTypes(Exception):
    """Geometry comparison across geometry types is meaningless"""
    pass


def clean(geom):
    """Clean invalid geometries using ``buffer(0)`` trick.

    ``geom`` is a shapely geometry; returns a shapely geometry."""
    if not geom.is_valid:
        geom = geom.buffer(0)
    return geom if geom.is_valid else GeometryCollection([])


def recursive_geom_finder(geom, kind):
    """Return all elements of ``geom`` that are of ``kind``. For example, return all linestrings in a geometry collection.

    ``geom`` is a Shapely geometry.

    ``kind`` should be one of ``("line", "point", "polygon")``.

    Returns either a ``MultiPoint``, ``MultiLineString``, or ``MultiPolygon``. Returns ``None`` is no valid element is found."""
    assert kind in ("line", "point", "polygon"), "Invalid ``kind``"

    TYPES = {
        'line': (LineString, LinearRing, MultiLineString),
        'point': (Point, MultiPoint),
        'polygon': (Polygon, MultiPolygon),
    }
    CONTAINER = {
        'line': MultiLineString,
        'point': MultiPoint,
        'polygon': MultiPolygon,
    }

    def recurse(geom, types):
        if isinstance(geom, types):
            yield geom
        elif isinstance(geom, GeometryCollection):
            for elem in geom:
                for val in recurse(elem, types):
                    yield val
        else:
            yield None

    elements = [elem for elem in recurse(geom, TYPES[kind])
                if elem is not None]
    if not elements:
        return None
    else:
        geom = clean(cascaded_union(elements))
        if 'Multi' not in geom.type:
            geom = CONTAINER[kind]([geom])
        return geom


def get_intersection(obj, kind, collection, indices,
                     to_meters=True,
                     return_geoms=True):
    """Return a dictionary describing the intersection of ``obj`` with ``collection[indices]``.

    ``obj`` is a Shapely geometry.
    ``kind`` is one of ``("line", "point", "polygon")`` - the kind of object to be returned.
    ``collection`` is a ``Map``.
    ``indices`` is an iterator of integers; indices into ``collection``.
    ``projection_func`` is a function to project the results to a new CRS before taking area, etc. If falsey, no projection will take place.
    ``return_geoms``: Return intersected geometries in addition to area, etc.

    Assumes that the polygons in ``collection`` do not overlap.

    Returns a dictionary of form:

    .. code-block:: python

        {
            collection_index: {
                'measure': measure of are or length,
                'geom': intersected geometry # if return_geoms
            }
        }

    The algorithm used for line and point intersections is incorrect - it will double count lines which lay along the borders of two polygons, and point that lie on the border of two polygons. A more robust function would take substantially more development and computation time, and total error should be less than 10 percent.

    """
    assert kind in ("line", "point", "polygon"), "Invalid ``kind``"

    proj_func = project if to_meters else lambda x: x
    obj = clean(obj)

    results = {}

    for index, geom in collection.iter_latlong(indices):
        if not geom.intersects(obj):
            continue
        g = recursive_geom_finder(
            clean(obj.intersection(geom)),
            kind
        )
        if not g:
            continue
        results[index] = {'measure': get_measure(proj_func(g), kind)}
        if return_geoms:
            results[index]['geom'] = g

    return results


def get_measure(geom, kind=None):
    """Get area, length, or number of points in ``geom``.

    * ``geom``: A shapely geom.
    * ``kind``: Geometry type, optional. One of `polygon`, `line`, or `point`.

    Kind will be guessed based on type of ``geom`` if not otherwise provided.

    If ``kind`` is not one of the allowed types, raises ``ValueError``.

    Returns a float."""
    if kind is None:
        kind = kind_mapping.get(geom.geom_type)

    if kind == 'polygon':
        return geom.area
    elif kind == 'line':
        return geom.length
    elif kind == 'point':
        if geom.geom_type == 'MultiPoint':
            return float(len(geom))
        elif geom.geom_type == 'Point':
            return 1.
    raise ValueError(
        "No applicable measure for geom: {}".format(geom)
    )


def get_remaining(original, geoms, to_meters=True):
    """Get the remaining area/length/number from ``original`` after subtracting the union of ``geoms``.

    * ``original``: Shapely geom in WGS84 CRS.
    * ``geoms``: List of shapely geoms in WGS84 CRS.
    * ``to_meters``: Boolean. Return value calculated in Mollweide projection.

    ``original`` and ``geoms`` should have the same geometry type, and ``geoms`` are components of ``original``.

    Returns a float."""
    try:
        kind = kind_mapping[original.geom_type]
    except KeyError:
        raise ValueError(
            "Can't use this geometry type: {}".format(original.geom_type)
        )

    if not to_meters or kind == 'point':
        proj_func = lambda x: x
    else:
        proj_func = project

    if geoms and {kind_mapping[g.geom_type] for g in geoms} != {kind}:
        raise IncompatibleTypes

    actual = get_measure(proj_func(original))
    if geoms:
        union_total = get_measure(proj_func(cascaded_union(geoms)), kind)
        individ_total = sum(get_measure(proj_func(geom), kind) for geom in geoms)
        return (actual - union_total) * (individ_total / union_total)
    else:
        return actual



# -*- coding: utf-8 -*-
from functools import partial
from shapely.ops import transform
import pyproj


WGS84 = "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs"
# See also http://spatialreference.org/ref/esri/54009/
# and http://cegis.usgs.gov/projection/pdf/nmdrs.usery.prn.pdf
MOLLWEIDE = "+proj=moll +lon_0=0 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs"


def wgs84(s):
    """Fix no CRS or fiona giving abbreviated wgs84 definition.

    Returns WGS84 if ``s`` is falsey."""
    if s == "+no_defs" or not s:
        return WGS84
    else:
        return s


def project(geom, from_proj=None, to_proj=None):
    """
Project a ``shapely`` geometry, and returns a new geometry of the same type from the transformed coordinates.

Default input projection is `WGS84 <https://en.wikipedia.org/wiki/World_Geodetic_System>`_, default output projection is `Mollweide <http://spatialreference.org/ref/esri/54009/>`_.

Inputs:
    *geom*: A ``shapely`` geometry.
    *from_proj*: A ``PROJ4`` string. Optional.
    *to_proj*: A ``PROJ4`` string. Optional.

Returns:
    A ``shapely`` geometry.

    """
    from_proj = wgs84(from_proj)
    if to_proj is None:
        to_proj = MOLLWEIDE
    else:
        to_proj = wgs84(to_proj)

    to_proj, from_proj = pyproj.Proj(to_proj), pyproj.Proj(from_proj)

    if ((to_proj == from_proj) or
            (to_proj.is_latlong() and from_proj.is_latlong())):
        return geom

    projection_func = partial(pyproj.transform, from_proj, to_proj)
    return transform(projection_func, geom)



# -*- coding: utf-8 -*-
from .conversion import check_type
from .filesystem import sha256
from .projection import project
from fiona import crs as fiona_crs
from functools import partial
from shapely.geometry import shape
import fiona
import os
import rtree


class DuplicateFieldID(Exception):
    """Field ID value is duplicated and should be unique"""
    pass


class Map(object):
    """A wrapper around fiona ``open`` that provides some additional functionality.

    Requires an absolute filepath.

    Additional metadata can be provided in `kwargs`:
        * `layer` specifies the shapefile layer

    .. warning:: The Fiona field ``id`` is not used, as there are no real constraints on these values or values types (see `Fiona manual <http://toblerity.org/fiona/manual.html#record-id>`_), and real world data is often dirty and inconsistent. Instead, we use ``enumerate`` and integer indices.

    """
    def __init__(self, filepath, identifying_field=None, **kwargs):
        assert os.path.exists(filepath), "No file at given path"

        self.filepath = filepath
        self.fieldname = identifying_field
        self.metadata = kwargs

        assert check_type(filepath) == 'vector', \
            "Must give a vector dataset"

        with fiona.drivers():
            self.file = fiona.open(
                self.filepath,
                **kwargs
            )

    def iter_latlong(self, indices=None):
        """Iterate over dataset as Shapely geometries in WGS 84 CRS."""
        _ = partial(project, from_proj=self.crs, to_proj='')
        if indices is None:
            for index, feature in enumerate(self):
                yield (index, _(shape(feature['geometry'])))
        else:
            for index in indices:
                yield (index, _(shape(self[index]['geometry'])))

    def create_rtree_index(self):
        """Create `rtree <http://toblerity.org/rtree/>`_ index for efficient spatial querying.

        **Note**: Bounds are given in lat/long, not in the native CRS"""
        self.rtree_index = rtree.Rtree()
        for index, geom in self.iter_latlong():
            self.rtree_index.add(index, geom.bounds)
        return self.rtree_index

    def get_fieldnames_dictionary(self, fieldname=None):
        fieldname = fieldname or self.fieldname
        assert fieldname, "No valid identifying field name"
        assert fieldname in next(iter(self.file))['properties'], \
            "Given fieldname not in file"
        fd = {index: obj['properties'].get(fieldname, None) \
            for index, obj in enumerate(self)}
        if len(fd.keys()) != len(set(fd.values())):
            raise DuplicateFieldID(
                "Given field name not unique for all records"
            )
        return fd

    @property
    def geometry(self):
        geom = self.file.meta['schema']['geometry']
        if geom == 'Unknown':
            geoms = {obj['geometry']['type'] for obj in self}
            if len(geoms) == 1:
                return geoms.pop()
            else:
                return 'Unknown'
        return geom

    @property
    def hash(self):
        return sha256(self.filepath)

    @property
    def crs(self):
        """Coordinate reference system, as defined by vector file."""
        return fiona_crs.to_string(self.file.crs)

    def __iter__(self):
        return iter(self.file)

    def _create_index_map(self):
        self._index_map = {index: int(feature['id']) for index, feature in enumerate(self)}

    def __getitem__(self, index):
        """Get feature from a fiona dataset.

        As Fiona is just a `simple wrapper to GDAL <https://github.com/Toblerity/Fiona/blob/0af2eac3fdee25d9660e4d90dcf97513cb4a15b4/fiona/ogrext2.pyx#L715>`__, and `GDAL has no guarantees on index starting values or continuity <https://trac.osgeo.org/gdal/ticket/356>`__, we construct a mapping dictionary from what we get when we enumerate the source file to what Python expects. This mapping dictionary is only created the first time ``__getitem__`` is called. Among commonly used formats, only Geopackage starts with 1 (geopackage[0] will just return ``None``).

        Note that our lookup dictionary breaks negative indexing."""
        if not hasattr(self, "_index_map"):
            self._create_index_map()
        return self.file[self._index_map[index]]

    def __len__(self):
        return len(self.file)



from .filesystem import sha256, get_appdirs_path
import fiona
import os
import rasterio
import tempfile
import warnings

from rasterio.rio.helpers import coords, write_features
from rasterio.crs import CRS

import numpy as np
import rasterio.features
import rasterio.warp


def check_type(filepath):
    """Determine if a GIS dataset is raster or vector.

    ``filepath`` is a filepath of a GIS dataset file.

    Returns ``'vector'`` or ``'raster'``. Raises a ``ValueError`` if the file can't be opened with fiona or rasterio."""
    try:
        with fiona.open(filepath) as ds:
            assert ds.meta['schema']['geometry'] != 'None'
        return 'vector'
    except:
        try:
            with rasterio.open(filepath) as ds:
                assert ds.meta
            return 'raster'
        except:
            raise ValueError("Unknown data type")


def convert_to_vector(filepath, dirpath=None, band=1):
    """Convert raster file at ``filepath`` to a vector file. Returns filepath of created vector file.

    ``dirpath`` should be a writable directory. If ``dirpath`` is no specified, uses the `appdirs library <https://pypi.python.org/pypi/appdirs>`__ to find an appropriate directory.

    ``band`` should be the integer index of the band; default is 1. Note that band indices start from 1, not 0.

    The generated vector file will be in GeoJSON, and have the WGS84 CRS.

    Because we are using `GDAL polygonize <http://www.gdal.org/gdal__alg_8h.html#a7a789015334d677afcbef67e5a6b4a7c>`__, we can't use 64 bit floats. This function will automatically convert rasters from 64 to 32 bit floats if necessary."""
    assert isinstance(band, int), "band must be an integer"

    if dirpath is None:
        dirpath = get_appdirs_path("raster-conversion")
    else:
        assert (os.path.isdir(dirpath) and os.access(dirpath, os.W_OK)), \
            "dirpath must be a writable directory"

    out_fp = os.path.join(dirpath, "{}.{}.geojson".format(sha256(filepath), band))
    if os.path.exists(out_fp):
        return out_fp

    _shapes(filepath, out_fp, band)
    return out_fp


def _shapes(in_fp, out_fp, bidx):
    """Code from rio CLI: https://github.com/mapbox/rasterio/blob/master/rasterio/rio/shapes.py
    All the click stuff is cut out, as well as some option which we don't need.

    Extracts shapes from one band or mask of a dataset and writes
    them out as GeoJSON. Shapes will be transformed to WGS 84 coordinates.

    The default action of this command is to extract shapes from the
    first band of the input dataset. The shapes are polygons bounding
    contiguous regions (or features) of the same raster value. This
    command performs poorly for int16 or float type datasets."""

    dump_kwds = {'sort_keys': True}

    # This is the generator for (feature, bbox) pairs.
    class Collection(object):

        def __init__(self):
            self._xs = []
            self._ys = []

        @property
        def bbox(self):
            return min(self._xs), min(self._ys), max(self._xs), max(self._ys)

        def __call__(self):
            with rasterio.open(in_fp) as src:
                if bidx > src.count:
                    raise ValueError('bidx is out of range for raster')

                # Adjust transforms.
                transform = src.transform

                # Most of the time, we'll use the valid data mask.
                # We skip reading it if we're extracting every possible
                # feature (even invalid data features) from a band.
                msk = src.read_masks(bidx)
                img = src.read(bidx, masked=False)

                # Transform the raster bounds.
                bounds = src.bounds
                xs = [bounds[0], bounds[2]]
                ys = [bounds[1], bounds[3]]
                xs, ys = rasterio.warp.transform(
                    src.crs, CRS({'init': 'epsg:4326'}), xs, ys)
                self._xs = xs
                self._ys = ys

                # Prepare keyword arguments for shapes().
                kwargs = {'transform': transform}
                kwargs['mask'] = msk

                src_basename = os.path.basename(src.name)

                # Yield GeoJSON features.
                for i, (g, val) in enumerate(
                        rasterio.features.shapes(img, **kwargs)):
                    g = rasterio.warp.transform_geom(
                        src.crs, 'EPSG:4326', g,
                        antimeridian_cutting=True, precision=-1)
                    xs, ys = zip(*coords(g))
                    yield {
                        'type': 'Feature',
                        'id': str(i),
                        'properties': {
                            'val': val,
                            'filename': src_basename,
                            'id': i
                        },
                        'bbox': [min(xs), min(ys), max(xs), max(ys)],
                        'geometry': g
                    }

    with open(out_fp, "w") as f:
        write_features(
            f, Collection(), sequence=False,
            geojson_type='collection', use_rs=False, **dump_kwds
        )


def clean_raster(fp, new_fp=None, band=1, nodata=None):
    """Clean raster data and metadata:
        * Delete invalid block sizes, and remove tiling
        * Set nodata to a reasonable value, if possible
        * Convert to 32 bit floats, if currently 64 bit floats and such conversion is possible

    ``fp``: String. Filepath of the input raster file.

    ``new_fp``: String, optional. Filepath of the raster to create. If not provided, the new raster will have the same name as the existing file, but will be created in a temporary directory.

    ``band``: Integer, default is ``1``. Raster band to clean and create in new file. Each band of a multiband raster would have to be cleaned separately.

    ``nodata``: Float, optional. Additional value to try when changing ``nodata`` value; must not be present in existing raster data.

    Returns the filepath of the new file as a compressed GeoTIFF. Can also return ``None`` if no new raster was written due to failing preconditions."""
    with rasterio.Env():
        with rasterio.open(fp) as src:
            profile = src.profile
            array = src.read(band)
            dtypes = src.dtypes

    profile.update(
        driver='GTiff',
        count=1,
        compress='lzw')

    # Set nodata to a reasonable value if possible
    if profile.get('nodata') is None and (array < -1e30).sum():
        warnings.warn("No `nodata` value set, but large negative numbers present. "
              "Please set a valid `nodata` value in raster file.")
        return
    elif profile.get('nodata') and profile['nodata'] < -1e30:
        nodatas = [-1, -99, -999, -9999]
        if nodata is not None:
            nodatas = [nodata] + nodatas
        found = False
        for value in nodatas:
            if not (array == value).sum():
                nodata, found = value, True
                break

        if found:
            array[np.isclose(array, profile['nodata'])] = nodata
            array[np.isnan(array)] = nodata
            profile['nodata'] = nodata
        else:
            warnings.warn((
                "`nodata` value is large and negative ({}), but no suitable "
                "replacement value found. Please specify a `nodata` value."
            ).format(profile['nodata']))
            return

    if dtypes[band - 1] == rasterio.float64:
        if not ((array < np.finfo('float32').min).sum() or
                (array > np.finfo('float32').max).sum()):
            array = array.astype(np.float32)
            profile['dtype'] = np.float32
        else:
            print("Not converting to 32 bit float; out of range values present.")

    profile['tiled'] = False
    profile.pop('blockysize', None)
    profile.pop('blockxsize', None)

    if new_fp is None:
        new_fp = os.path.join(tempfile.mkdtemp(), os.path.basename(fp))

    with rasterio.Env():
        with rasterio.open(new_fp, 'w', **profile) as dst:
            dst.write(array, 1)

    return new_fp


def round_to_x_significant_digits(array, sf=3):
    num_digits = sf - np.floor(np.log10(np.abs(array))).astype(int) - 1
    num_digits[np.where(array == 0)] = 0
    for value in np.unique(num_digits):
        indices = np.where(num_digits == value)
        array[indices] = np.round(array[indices], value)
    return array


def round_raster(in_fp, out_fp=None, band=1, sig_digits=3):
    """Round raster cell values to a certain number of significant digits in new raster file. For example, π rounded to 4 significant digits is 3.142.

        * ``in_fp``: String. Filepath of raster input file.
        * ``out_fp``: String, optional. Filepath of new raster to be created. Should not currently exist. If not provided, the new raster will have the same name as the existing file, but will be created in a temporary directory.
        * ``band``: Int, default is 1. Band to round. Band indices start at 1.
        * ``sig_digits``: Int, default is 3. Number of significant digits to round to.

    The created raster file will have the same ``dtype``, shape, and CRS as the input file. It will be a compressed GeoTIFF.

    Returns ``out_fp``, the filepath of the created file.

    """
    if out_fp is None:
        out_fp = os.path.join(
            tempfile.mkdtemp(),
            os.path.basename(in_fp)
        )

    with rasterio.Env():
        with rasterio.open(in_fp) as src:
            array = src.read(band)
            profile = src.profile

        profile.update(
            driver='GTiff',
            count=1,
            compress='lzw')

        with rasterio.open(out_fp, 'w', **profile) as dst:
            dst.write(round_to_x_significant_digits(array, sig_digits), 1)

    return out_fp



import os
import appdirs
import bz2
import codecs
import hashlib
import json


def sha256(filepath, blocksize=65536):
    """Generate SHA 256 hash for file at ``filepath``.

    ``blocksize`` (default is 65536) is block size to feed to hasher.

    Returns a ``str``."""
    hasher = hashlib.sha256()
    fo = open(filepath, 'rb')
    buf = fo.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = fo.read(blocksize)
    return hasher.hexdigest()


def json_exporter(data, filepath, compress=True):
    """Export a file to JSON. Compressed with ``bz2`` is ``compress``.

    Returns the filepath of the JSON file. Returned filepath is not necessarily ``filepath``, if ``compress`` is ``True``."""
    if compress:
        filepath += ".bz2"
        with bz2.BZ2File(filepath, "w") as f:
            f.write(json.dumps(data, ensure_ascii=False).encode('utf-8'))
    else:
        with codecs.open(filepath, "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False)
    return filepath


def json_importer(fp):
    """Load a JSON file. Can be compressed with ``bz2`` - if so, it should have the extension ``.bz2``.

    Returns the data in the JSON file."""
    if fp.endswith(".bz2"):
        with bz2.open(fp, "rb") as f:
            data = json.loads(f.read().decode("utf-8"))
    else:
        with open(fp, "r") as f:
            data = json.load(f)
    return data


def get_appdirs_path(subdir):
    """Get path for an ``appdirs`` directory, with subdirectory ``subdir``.

    Returns the full directory path."""
    dirpath = os.path.join(
        appdirs.user_data_dir("pandarus", "pandarus-cache"),
        subdir
    )
    if not os.path.isdir(dirpath):
        os.makedirs(dirpath)
    return os.path.abspath(dirpath)



# -*- coding: utf-8 -*-
from .conversion import check_type
from .filesystem import json_exporter, get_appdirs_path, sha256, json_importer
from .maps import Map
from .intersections import intersection_dispatcher
from .geometry import get_remaining
from .projection import project
from .rasters import gen_zonal_stats
from fiona.crs import from_string
from functools import partial
from shapely.geometry import mapping, shape
import datetime
import fiona
import multiprocessing
import os
import rasterio
import warnings


WGS84 = from_string("+datum=WGS84 +ellps=WGS84 +no_defs +proj=longlat")

MISMATCHED_CRS = """Possible coordinate reference systems (CRS) mismatch. The raster statistics may be incorrect, please only use this method when both vector and raster have the same CRS.
    Vector: {}
    Raster: {}"""

CPU_COUNT = multiprocessing.cpu_count()


def get_map(fp, field, kwargs):
    obj = Map(fp, field, **kwargs)
    metadata = {
        'sha256': obj.hash,
        'filename': os.path.basename(fp),
        'field': field,
        'path': os.path.abspath(fp),
    }
    return obj, metadata


def raster_statistics(vector_fp, identifying_field, raster, output=None,
        band=1, compress=True, fiona_kwargs={}, **kwargs):
    """Create statistics by matching ``raster`` against each spatial unit in ``self.from_map``.

    For each spatial unit in ``self.from_map``, calculates the following statistics for values from ``raster``: min, mean, max, and count. Count is the number of raster cells intersecting the vector spatial unit. No data values in the raster are not including in the generated statistics.

    This function uses a fork of the ``rasterstats`` library that break each raster cell into 100 smaller cells, as a compromise approach to handle the fact that some raster cells are completely with a vector geometry, while others only have a small fraction of their cell area within the vector geometry. Each of the 100 small raster cells is weighted equally, and each is tested to make sure it intersects the vector geometry.

    This function assumes that each smaller raster cell has the same area. This may change in the future.

    Input parameters:

        * ``vector_fp``: str. Filepath of the vector dataset.
        * ``identifying_field``: str. Name of the field in ``vector_fp`` that uniquely identifies each feature.
        * ``raster``: str. Filepath of the raster dataset.
        * ``output``: str, optional. Filepath of the output file. Will be deleted if it exists already.
        * ``band``: int, optional. Raster band used for calculations. Default is ``1``.
        * ``compress``: bool, optional. Compress JSON results file. Default is ``True``.
        * ``fiona_kwargs``: dict, optional. Additional arguments to pass to fiona when opening ``vector_fp``.

    Any additional ``kwargs`` are passed to ``gen_zonal_stats``.

    Output format:

    Output is a (maybe compressed) JSON file with the following schema:

     .. code-block:: python

        {
            'metadata': {
                'vector': {
                    'field': 'name of uniquely identifying field',
                    'path': 'path to vector input file',
                    'sha256': 'sha256 hash of input file'
                },
                'raster': {
                    'band': 'band used to calculate raster stats',
                    'path': 'path to raster input file',
                    'filename': 'name of raster file',
                    'sha256': 'sha256 hash of input file'
                },
                'when': 'datetime this calculation finished, ISO format'
            },
            'data': [
                [
                    'vector `identifying_field` value',
                    {
                        'count': 'number of raster cells included. float because consider fractional intersections',
                        'min': 'minimum raster value in this vector feature',
                        'mean': 'average raster value in this vector feature',
                        'max': 'maximum raster value in this vector feature',
                    }
                ]
            ]
        }

    """
    vector, v_metadata = get_map(vector_fp, identifying_field, fiona_kwargs)
    assert check_type(raster) == 'raster'

    with rasterio.open(raster) as r:
        raster_crs = r.crs.to_string()
        meta = r.meta

    if vector.crs != raster_crs:
        warnings.warn(MISMATCHED_CRS.format(vector.crs, raster_crs))

    if not output:
        dirpath = get_appdirs_path("rasterstats")
        output = os.path.join(
            dirpath,
            "{}-{}-{}.json".format(vector.hash, sha256(raster), band)
        )

    if os.path.exists(output):
        os.remove(output)

    pcw = meta['height'] < 5000 and meta['width'] < 10000

    stats_generator = gen_zonal_stats(vector_fp, raster, band=band, percent_cover_weighting=pcw, **kwargs)
    mapping_dict = vector.get_fieldnames_dictionary()
    results = [(mapping_dict[index], row)
               for index, row in enumerate(stats_generator)]

    metadata = {
        'vector': v_metadata,
        'raster': {
            'sha256': sha256(raster),
            'path': raster,
            'filename': os.path.basename(raster),
            'band': band
        },
        'when': datetime.datetime.now().isoformat()
    }
    return json_exporter({'data': results, 'metadata': metadata}, output, compress)


def as_features(dct):
    for index, key in enumerate(dct):
        row = dct[key]
        gj = {
            'geometry': mapping(row['geom']),
            'properties': {
                'id': index,
                'from_label': key[0],
                'to_label': key[1],
                'measure': row['measure']},
        }
        yield gj


def intersect(first_fp, first_field, second_fp, second_field,
        first_kwargs={}, second_kwargs={}, dirpath=None, cpus=CPU_COUNT,
        driver='GeoJSON', compress=True, log_dir=None):
    """Calculate the intersection of two vector spatial datasets.

    The first spatial input file **must** have only one type of geometry, i.e. points, lines, or polygons, and excluding geometry collections. Any of the following are allowed: Point, MultiPoint, LineString, LinearRing, MultiLineString, Polygon, MultiPolygon.

    The second spatial input file **must** have either Polygons or MultiPolygons. Although no checks are made, this and other functions make a strong assumption that the spatial units in the second spatial unit do not overlap.

    Input parameters:

        * ``first_fp``: String. File path to the first spatial dataset.
        * ``first_field``: String. Name of field that uniquely identifies features in the first spatial dataset.
        * ``second_fp``: String. File path to the second spatial dataset.
        * ``second_field``: String. Name of field that uniquely identifies features in the second spatial dataset.
        * ``first_kwargs``: Dictionary, optional. Additional arguments, such as layer name, passed to fiona when opening the first spatial dataset.
        * ``second_kwargs``: Dictionary, optional. Additional arguments, such as layer name, passed to fiona when opening the second spatial dataset.
        * ``dirpath``: String, optional. Directory to save output files.
        * ``cpus``: Integer, default is ``multiprocessing.cpu_count()``. Number of CPU cores to use when calculating. Use ``cpus=0`` to avoid starting a multiprocessing pool.
        * ``driver``: String, default is ``GeoJSON``. Fiona driver name to use when writing geospatial output file. Common values are ``GeoJSON`` or ``GPKG``.
        * ``compress``: Boolean, default is True. Compress JSON output file.
        * ``log_dir``: String, optional.

    Returns filepaths for two created files.

    The first is a geospatial file that has the geometry of each possible intersection of spatial units from the two input files. The geometry type of this file will depend on the geometry type of the first input file, but will always be a multi geometry, i.e. one of MultiPoint, MultiLineString, MultiPolygon. This file will also always have the `WGS 84 CRS <http://spatialreference.org/ref/epsg/wgs-84/>`__. The output file has the following schema:

        * ``id``: Integer. Auto-increment field starting from zero.
        * ``from_label``: String. The value for the uniquely identifying field from the first input file.
        * ``to_label``: String. The value for the uniquely identifying field from the second input file.
        * ``measure``: Float. A measure of the intersected shape. For polygons, this is the area of the feature in square meters. For lines, this is the length in meters. For points, this is the number of points. Area and length calculations are made using the `Mollweide projection <https://en.wikipedia.org/wiki/Mollweide_projection>`__.

    The second file is an extract of some of the feature fields in the JSON data format. This is used by programs that don't need to depend on GIS data libraries. The JSON format is:

    .. code-block:: python

        {
            'metadata': {
                'first': {
                    'field': 'name of uniquely identifying field',
                    'path': 'path to first input file',
                    'filename': 'name of first input file',
                    'sha256': 'sha256 hash of input file'
                },
                'second': {
                    'field': 'name of uniquely identifying field',
                    'path': 'path to second input file',
                    'filename': 'name of second input file',
                    'sha256': 'sha256 hash of input file'
                },
                'when': 'datetime this calculation finished, ISO format'
            },
            'data': [
                [
                    'identifying field for first file',
                    'identifying field for second file',
                    'measure value'
                ]
            ]
        }

    """
    first, first_metadata = get_map(first_fp, first_field, first_kwargs)
    second, second_metadata = get_map(second_fp, second_field, second_kwargs)

    if not dirpath:
        dirpath = get_appdirs_path("intersections")

    base_filepath = os.path.join(dirpath, "{}.{}.".format(
        first.hash, second.hash
    ))

    fiona_fp = base_filepath + driver.lower()
    data_fp = base_filepath + "json"

    if os.path.exists(fiona_fp):
        os.remove(fiona_fp)
    if os.path.exists(data_fp):
        os.remove(data_fp)

    data = intersection_dispatcher(first_fp, second_fp, cpus=cpus, log_dir=log_dir)

    first_mapping = first.get_fieldnames_dictionary()
    second_mapping = second.get_fieldnames_dictionary()
    data = {
        (first_mapping[k[0]], second_mapping[k[1]]): v
        for k, v in data.items()
    }

    schema = {
        'properties': {
            'id': 'int',
            'from_label': first.file.meta['schema']['properties'][first_field],
            'to_label': second.file.meta['schema']['properties'][second_field],
            'measure': 'float',
        },
        'geometry': 'MultiPolygon',
    }

    with fiona.drivers():
        with fiona.open(
                fiona_fp, 'w',
                crs=WGS84,
                driver=driver,
                schema=schema,
            ) as sink:
            for f in as_features(data):
                sink.write(f)

    data_fp = json_exporter(
        {
            'data': [(k[0], k[1], v['measure']) for k, v in data.items()],
            'metadata':
                {
                    'first': first_metadata,
                    'second': second_metadata,
                    'when': datetime.datetime.now().isoformat(),
                }
        },
        data_fp,
        compress=compress
    )

    return fiona_fp, data_fp


def calculate_remaining(source_fp, source_field, intersection_fp,
        source_kwargs={}, dirpath=None, compress=True):
    """Calculate the remaining area/length/number of points left out of an intersections file generated by ``intersect``.

    Input parameters:

        * ``source_fp``: String. Filepath of the input spatial data which could have features outside of the intersection result.
        * ``source_field``: String. Name of field that uniquely identifies features in the input spatial dataset.
        * ``intersection_fp``: Filepath of the intersection spatial dataset generated by the ``intersect`` function.
        * ``source_kwargs``: Dictionary, optional. Additional arguments, such as layer name, passed to fiona when opening the input spatial dataset.
        * ``dirpath``: String, optional. Directory where the output file will be saved.
        * ``compress``: Boolean. Whether or not to compress the output file.

    .. warning:: ``source_fp`` must be the first file provided to the ``intersect`` function, **not** the second!

    Returns the filepath of the output file. The output file JSON format is:

    .. code-block:: python

        {
            'metadata': {
                'source': {
                    'field': 'name of uniquely identifying field',
                    'path': 'path to the input file',
                    'filename': 'name of the input file',
                    'sha256': 'sha256 hash of the input file'
                },
                'intersections': {
                    'field': 'name of uniquely identifying field (always `id`)',
                    'path': 'path to intersections spatial dataset',
                    'filename': 'name of intersections spatial dataset',
                    'sha256': 'sha256 hash of intersection spatial dataset'
                }
                'when': 'datetime this calculation finished, ISO format'
            },
            'data': [
                [
                    'identifying field for source file',
                    'measure value'
                ]
            ]
        }

    """
    source, source_metadata = get_map(source_fp, source_field, source_kwargs)
    intersections, inter_metadata = get_map(intersection_fp, 'id', {})

    assert intersections.file.schema['properties'].keys() == \
        {'id', 'from_label', 'to_label', 'measure'}
    assert intersections.file.schema['properties']['id'] == 'int'
    assert intersections.file.schema['properties']['measure'] == 'float'

    if not dirpath:
        dirpath = get_appdirs_path("intersections")

    output = os.path.join(dirpath, "{}.{}.json".format(source.hash, intersections.hash)
    )

    _ = partial(project, from_proj=source.crs, to_proj='')

    def get_geoms(feat):
        return [
            shape(x['geometry'])
            for x in intersections
            if (x['properties']['from_label'] ==
                feat['properties'][source_field])
        ]

    data = [(
            feat['properties'][source_field],
            get_remaining(
                _(shape(feat['geometry'])), get_geoms(feat)
            )
        ) for feat in source
    ]

    metadata = {
        'source': source_metadata,
        'intersections': inter_metadata,
        'when': datetime.datetime.now().isoformat(),
    }

    return json_exporter({'data': data, 'metadata': metadata}, output, compress)


def intersections_from_intersection(fp, metadata=None, dirpath=None):
    """Process an intersections spatial dataset to create two intersections data files.

    ``fp`` is the file path of a vector dataset created by the ``intersect`` function. The intersection of two spatial scales (A, B) is a third spatial scale (C); this function creates intersection data files for (A, C) and (B, C).

    As the intersections data file includes metadata on the input files, this function must have access to the intersections data file created at the same time as intersections spatial dataset. If the ``metadata`` filepath is not provided, the metadata file is looked for in the same directory as ``fp``.

    Returns the file paths of the two new intersections data files.
    """
    assert os.path.isfile(fp)

    if metadata:
        assert os.path.isfile(metadata)
    else:
        metadata = ".".join(fp.split(".")[:-1]) + ".json"
        if not os.path.isfile(metadata):
            metadata += ".bz2"
            if not os.path.isfile(metadata):
                raise ValueError("Can't find metadata file")

    metadata = json_importer(metadata)['metadata']

    with fiona.open(fp) as source:
        for key in ('id', 'from_label', 'to_label', 'measure'):
            assert key in source.schema['properties']
        data = [feat['properties'] for feat in source]

    this = {
        'field': 'id',
        'path': fp,
        'filename': os.path.basename(fp),
        'sha256': sha256(fp)
    }

    first_dataset = {
        'data': [(o['id'], o['from_label'], o['measure']) for o in data],
        'metadata': {
            'first': this,
            'second': metadata['first'],
            'when': datetime.datetime.now().isoformat()
        }
    }
    second_dataset = {
        'data': [(o['id'], o['to_label'], o['measure']) for o in data],
        'metadata': {
            'first': this,
            'second': metadata['second'],
            'when': datetime.datetime.now().isoformat()
        }
    }

    if not dirpath:
        dirpath = get_appdirs_path("intersections")

    first_fp = os.path.join(
        dirpath,
        "{}.{}.json".format(this['sha256'], metadata['first']['sha256'])
    )
    second_fp = os.path.join(
        dirpath,
        "{}.{}.json".format(this['sha256'], metadata['second']['sha256'])
    )

    return (
        json_exporter(first_dataset, first_fp),
        json_exporter(second_dataset, second_fp)
    )



# -*- coding: utf-8 -*-
import numpy as np
import warnings
from affine import Affine
from numpy import min_scalar_type
from rasterio import features
from rasterstats.io import read_features, Raster
from rasterstats.utils import (get_percentile, check_stats, remap_categories,
                    key_assoc_val, boxify_points)
from shapely.geometry import shape

# Code from https://github.com/perrygeo/python-rasterstats/pull/136
# Percent coverage selection and weighting
# Same license as rasterstats


def gen_zonal_stats(
        vectors, raster,
        layer=0,
        band=1,
        nodata=None,
        affine=None,
        stats=None,
        all_touched=True,
        percent_cover_selection=None,
        percent_cover_weighting=True,
        percent_cover_scale=20,
        categorical=False,
        category_map=None,
        add_stats=None,
        zone_func=None,
        raster_out=False,
        prefix=None,
        geojson_out=False, **kwargs):
    """Zonal statistics of raster values aggregated to vector geometries.

    Parameters

    vectors: path to an vector source or geo-like python objects

    raster: ndarray or path to a GDAL raster source
        If ndarray is passed, the ``affine`` kwarg is required.

    layer: int or string, optional
        If `vectors` is a path to an fiona source,
        specify the vector layer to use either by name or number.
        defaults to 0

    band: int, optional
        If `raster` is a GDAL source, the band number to use (counting from 1).
        defaults to 1.

    nodata: float, optional
        If `raster` is a GDAL source, this value overrides any NODATA value
        specified in the file's metadata.
        If `None`, the file's metadata's NODATA value (if any) will be used.
        defaults to `None`.

    affine: Affine instance
        required only for ndarrays, otherwise it is read from src

    stats:  list of str, or space-delimited str, optional
        Which statistics to calculate for each zone.
        All possible choices are listed in ``utils.VALID_STATS``.
        defaults to ``DEFAULT_STATS``, a subset of these.

    all_touched: bool, optional
        Whether to include every raster cell touched by a geometry, or only
        those having a center point within the polygon.
        defaults to `False`

    percent_cover_selection: float, optional
        Include only raster cells that have at least the given percent
        covered by the vector feature. Requires percent_cover_scale argument
        be used to specify scale at which to generate percent coverage
        estimates

    percent_cover_weighting: bool, optional
        whether or not to use percent coverage of cells during calculations
        to adjust stats (only applies to mean, count and sum)

    percent_cover_scale: int, optional
        Scale used when generating percent coverage estimates of each
        raster cell by vector feature. Percent coverage is generated by
        rasterizing the feature at a finer resolution than the raster
        (based on percent_cover_scale value) then using a summation to aggregate
        to the raster resolution and dividing by the square of percent_cover_scale
        to get percent coverage value for each cell. Increasing percent_cover_scale
        will increase the accuracy of percent coverage values; three orders
        magnitude finer resolution (percent_cover_scale=1000) is usually enough to
        get coverage estimates with <1% error in individual edge cells coverage
        estimates, though much smaller values (e.g., percent_cover_scale=10) are often
        sufficient (<10% error) and require less memory.

    categorical: bool, optional

    category_map: dict
        A dictionary mapping raster values to human-readable categorical names.
        Only applies when categorical is True

    add_stats: dict
        with names and functions of additional stats to compute, optional

    zone_func: callable
        function to apply to zone ndarray prior to computing stats

    raster_out: boolean
        Include the masked numpy array for each feature?, optional

        Each feature dictionary will have the following additional keys:
        mini_raster_array: The clipped and masked numpy array
        mini_raster_affine: Affine transformation
        mini_raster_nodata: NoData Value

    prefix: string
        add a prefix to the keys (default: None)

    geojson_out: boolean
        Return list of GeoJSON-like features (default: False)
        Original feature geometry and properties will be retained
        with zonal stats appended as additional properties.
        Use with `prefix` to ensure unique and meaningful property names.

    Returns

    generator of dicts (if geojson_out is False)
        Each item corresponds to a single vector feature and
        contains keys for each of the specified stats.

    generator of geojson features (if geojson_out is True)
        GeoJSON-like Feature as python dict
    """
    stats, run_count = check_stats(stats, categorical)

    # check inputs related to percent coverage
    percent_cover = False
    if percent_cover_weighting or percent_cover_selection is not None:
        percent_cover = True
        if percent_cover_scale is None:
            warnings.warn('No value for `percent_cover_scale` was given. '
                          'Using default value of 10.')
            percent_cover_scale = 10

        try:
            if percent_cover_scale != int(percent_cover_scale):
                warnings.warn('Value for `percent_cover_scale` given ({0}) '
                              'was converted to int ({1}) but does not '
                              'match original value'.format(
                                percent_cover_scale, int(percent_cover_scale)))

            percent_cover_scale = int(percent_cover_scale)

            if percent_cover_scale <= 1:
                raise Exception('Value for `percent_cover_scale` must be '
                                'greater than one ({0})'.format(
                                    percent_cover_scale))

        except:
            raise Exception('Invalid value for `percent_cover_scale` '
                            'provided ({0}). Must be type int.'.format(
                                percent_cover_scale))

        if percent_cover_selection is not None:
            try:
                percent_cover_selection = float(percent_cover_selection)
            except:
                raise Exception('Invalid value for `percent_cover_selection` '
                                'provided ({0}). Must be able to be converted '
                                'to a float.'.format(percent_cover_selection))

        # if not all_touched:
        #     warnings.warn('`all_touched` was not enabled but an option requiring '
        #                   'percent_cover calculations was selected. Automatically '
        #                   'enabling `all_touched`.')
        # all_touched = True

    with Raster(raster, affine, nodata, band) as rast:
        features_iter = read_features(vectors, layer)
        for _, feat in enumerate(features_iter):
            geom = shape(feat['geometry'])

            if 'Point' in geom.type:
                geom = boxify_points(geom, rast)
                percent_cover = False

            geom_bounds = tuple(geom.bounds)
            fsrc = rast.read(bounds=geom_bounds)

            if percent_cover:
                cover_weights = rasterize_pctcover_geom(
                    geom, shape=fsrc.shape, affine=fsrc.affine,
                    scale=percent_cover_scale,
                    all_touched=all_touched)
                rv_array = cover_weights > (percent_cover_selection or 0)
            else:
                rv_array = rasterize_geom(
                    geom, shape=fsrc.shape, affine=fsrc.affine,
                    all_touched=all_touched)

            # nodata mask
            isnodata = (fsrc.array == fsrc.nodata)

            # add nan mask (if necessary)
            if np.issubdtype(fsrc.array.dtype, float) and \
               np.isnan(fsrc.array.min()):
                isnodata = (isnodata | np.isnan(fsrc.array))

            # Mask the source data array
            # mask everything that is not a valid value or not within our geom
            masked = np.ma.MaskedArray(
                fsrc.array,
                mask=(isnodata | ~rv_array))

            # execute zone_func on masked zone ndarray
            if zone_func is not None:
                if not callable(zone_func):
                    raise TypeError(('zone_func must be a callable '
                                     'which accepts function a '
                                     'single `zone_array` arg.'))
                zone_func(masked)

            if masked.compressed().size == 0:
                # nothing here, fill with None and move on
                feature_stats = dict([(stat, None) for stat in stats])
                if 'count' in stats:  # special case, zero makes sense here
                    feature_stats['count'] = 0
            else:
                if run_count:
                    keys, counts = np.unique(masked.compressed(), return_counts=True)
                    pixel_count = dict(zip([np.asscalar(k) for k in keys],
                                           [np.asscalar(c) for c in counts]))

                if categorical:
                    feature_stats = dict(pixel_count)
                    if category_map:
                        feature_stats = remap_categories(category_map, feature_stats)
                else:
                    feature_stats = {}

                if 'min' in stats:
                    feature_stats['min'] = float(masked.min())
                if 'max' in stats:
                    feature_stats['max'] = float(masked.max())
                if 'mean' in stats:
                    if percent_cover:
                        feature_stats['mean'] = float(
                            np.sum(masked * cover_weights) /
                            np.sum(~masked.mask * cover_weights))
                    else:
                        feature_stats['mean'] = float(masked.mean())
                if 'count' in stats:
                    if percent_cover:
                        feature_stats['count'] = float(np.sum(~masked.mask * cover_weights))
                    else:
                        feature_stats['count'] = int(masked.count())
                # optional
                if 'sum' in stats:
                    if percent_cover:
                        feature_stats['sum'] = float(np.sum(masked * cover_weights))
                    else:
                        feature_stats['sum'] = float(masked.sum())
                if 'std' in stats:
                    feature_stats['std'] = float(masked.std())
                if 'median' in stats:
                    feature_stats['median'] = float(np.median(masked.compressed()))
                if 'majority' in stats:
                    feature_stats['majority'] = float(key_assoc_val(pixel_count, max))
                if 'minority' in stats:
                    feature_stats['minority'] = float(key_assoc_val(pixel_count, min))
                if 'unique' in stats:
                    feature_stats['unique'] = len(list(pixel_count.keys()))
                if 'range' in stats:
                    try:
                        rmin = feature_stats['min']
                    except KeyError:
                        rmin = float(masked.min())
                    try:
                        rmax = feature_stats['max']
                    except KeyError:
                        rmax = float(masked.max())
                    feature_stats['range'] = rmax - rmin

                for pctile in [s for s in stats if s.startswith('percentile_')]:
                    q = get_percentile(pctile)
                    pctarr = masked.compressed()
                    feature_stats[pctile] = np.percentile(pctarr, q)

            if 'nodata' in stats:
                featmasked = np.ma.MaskedArray(fsrc.array, mask=np.logical_not(rv_array))
                feature_stats['nodata'] = float((featmasked == fsrc.nodata).sum())

            if add_stats is not None:
                for stat_name, stat_func in add_stats.items():
                    feature_stats[stat_name] = stat_func(masked)

            if raster_out:
                feature_stats['mini_raster_array'] = masked
                feature_stats['mini_raster_affine'] = fsrc.affine
                feature_stats['mini_raster_nodata'] = fsrc.nodata

            if prefix is not None:
                prefixed_feature_stats = {}
                for key, val in feature_stats.items():
                    newkey = "{}{}".format(prefix, key)
                    prefixed_feature_stats[newkey] = val
                feature_stats = prefixed_feature_stats

            if geojson_out:
                for key, val in feature_stats.items():
                    if 'properties' not in feat:
                        feat['properties'] = {}
                    feat['properties'][key] = val
                yield feat
            else:
                yield feature_stats



def rasterize_geom(geom, shape, affine, all_touched=False):
    """
    Parameters
    ----------
    geom: GeoJSON geometry
    shape: desired shape
    affine: desired transform
    all_touched: rasterization strategy

    Returns
    -------
    ndarray: boolean
    """
    geoms = [(geom, 1)]
    rv_array = features.rasterize(
        geoms,
        out_shape=shape,
        transform=affine,
        fill=0,
        dtype='uint8',
        all_touched=all_touched)

    return rv_array.astype(bool)


# https://stackoverflow.com/questions/8090229/
#   resize-with-averaging-or-rebin-a-numpy-2d-array/8090605#8090605
def rebin_sum(a, shape, dtype):
    sh = shape[0],a.shape[0]//shape[0],shape[1],a.shape[1]//shape[1]
    return a.reshape(sh).sum(-1, dtype=dtype).sum(1, dtype=dtype)


def rasterize_pctcover_geom(geom, shape, affine, scale=None, all_touched=False):
    """
    Parameters
    ----------
    geom: GeoJSON geometry
    shape: desired shape
    affine: desired transform
    scale: scale at which to generate percent cover estimate

    Returns
    -------
    ndarray: float32
    """
    if scale is None:
        scale = 10

    min_dtype = min_scalar_type(scale ** 2)

    new_affine = Affine(affine[0]/scale, 0, affine[2],
                        0, affine[4]/scale, affine[5])

    new_shape = (shape[0] * scale, shape[1] * scale)

    rv_array = rasterize_geom(geom, new_shape, new_affine, all_touched=all_touched)

    rv_array = rebin_sum(rv_array, shape, min_dtype)

    return rv_array.astype('float32') / (scale**2)



__version__ = (1, 0, 3)

__all__ = (
    "Map",
    'clean_raster',
    'convert_to_vector',
    'calculate_remaining',
    'intersect',
    'intersections_from_intersection',
    'raster_statistics',
    'round_raster',
)

from .calculate import (
    calculate_remaining,
    intersect,
    intersections_from_intersection,
    raster_statistics,
)
from .maps import Map
from .conversion import convert_to_vector, clean_raster, round_raster



# -*- coding: utf-8 -*-
from .maps import Map
from .geometry import (
    clean,
    get_intersection,
    kind_mapping,
)
from .projection import project
from logging.handlers import QueueHandler, QueueListener
from shapely.geometry import shape
from shapely.geos import TopologicalError
import datetime
import logging
import math
import multiprocessing
import os


def chunker(iterable, chunk_size):
    for i in range(0, len(iterable), chunk_size):
        yield list(iterable[i:i + chunk_size])


def logger_init(dirpath=None):
    # Adapted from http://stackoverflow.com/a/34964369/164864
    logging_queue = multiprocessing.Queue()
    # this is the handler for all log records
    filepath = "{}-{}.log".format(
        'pandarus-worker', datetime.datetime.now().strftime("%d-%B-%Y-%I-%M%p")
    )
    if dirpath is not None:
        filepath = os.path.join(dirpath, filepath)
    handler = logging.FileHandler(
        filepath,
        encoding='utf-8',
    )
    handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(lineno)d %(message)s"))

    # queue_listener gets records from the queue and sends them to the handler
    queue_listener = QueueListener(logging_queue, handler)
    queue_listener.start()

    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    logger.addHandler(handler)

    return queue_listener, logging_queue


def worker_init(logging_queue):
    # Needed to pass logging messages from child processes to a queue
    # handler which in turn passes them onto queue listener
    queue_handler = QueueHandler(logging_queue)
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    logger.addHandler(queue_handler)


def get_jobs(map_size):
    # Want a reasonable chunk size
    # But also want a maximum of 200 jobs
    # Both numbers picked more or less at random...
    chunk_size = int(max(20, map_size / 200))
    num_jobs = int(math.ceil(map_size / float(chunk_size)))
    return chunk_size, num_jobs


def intersection_worker(from_map, from_objs, to_map, worker_id=1):
    """Multiprocessing worker for map matching"""
    logging.info("""Starting intersection_worker:
    from map: {}
    from objs: {} ({} to {})
    to map: {}
    worker id: {}""".format(from_map, len(from_objs or []) or 'all',
                            min(from_objs or [0]), max(from_objs or [0]),
                            to_map, worker_id))

    results = {}

    to_map = Map(to_map)
    if to_map.geometry not in ('Polygon', 'MultiPolygon'):
        raise ValueError("`to_map` geometry must be polygons")
    rtree_index = to_map.create_rtree_index()

    logging.info("Worker {}: Loaded `to` map.".format(worker_id))

    from_map = Map(from_map)
    try:
        kind = kind_mapping[from_map.geometry]
    except KeyError:
        raise ValueError("No valid geometry type in map {}".format(from_map))

    logging.info("Worker {}: Loaded `from` map.".format(worker_id))

    to_shape = lambda x: project(shape(x['geometry']), from_map.crs, '')

    if from_objs:
        from_gen = ((index, from_map[index]) for index in from_objs)
    else:
        from_gen = enumerate(from_map)

    for from_index, from_obj in from_gen:
        try:
            geom = clean(to_shape(from_obj))

            for k, v in get_intersection(
                geom,
                kind,
                to_map,
                rtree_index.intersection(geom.bounds),
                project
            ).items():
                results[(from_index, k)] = v

        except TopologicalError:
            logging.exception("Skipping topological error.")
            continue
        except:
            logging.exception("Intersection worker failed.")
            raise

    return results


def intersection_dispatcher(from_map, to_map, from_objs=None, cpus=None, log_dir=None):
    if not cpus:
        return intersection_worker(from_map, None, to_map)

    if from_objs:
        map_size = len(from_objs)
        ids = from_objs
    else:
        map_size = len(Map(from_map))
        ids = range(map_size)

    chunk_size, num_jobs = get_jobs(map_size)

    queue_listener, logging_queue = logger_init(log_dir)
    logging.info("""Starting `intersect` calculation.
    From map: {}
    To map: {}
    Map size: {}
    Chunk size: {}
    Number of jobs: {}""".format(
        from_map, to_map, map_size, chunk_size, num_jobs
    ))

    results = {}

    def callback_func(data):
        results.update(data)

    with multiprocessing.Pool(
                cpus or multiprocessing.cpu_count(),
                worker_init,
                [logging_queue]
            ) as pool:
        arguments = [
            (from_map, chunk, to_map, index)
            for index, chunk in enumerate(chunker(ids, chunk_size))
        ]

        function_results = []

        for argument_set in arguments:
            function_results.append(pool.apply_async(
                intersection_worker,
                argument_set,
                callback=callback_func
            ))
        for fr in function_results:
            fr.wait()

        if any(not fr.successful() for fr in function_results):
            raise ValueError("Couldn't complete Pandarus task")

    queue_listener.stop()

    logging.info("""Finished `intersect` calculation.
    From map: {}
    To map: {}
    Map size: {}
    Chunk size: {}
    Number of jobs: {}""".format(
        from_map, to_map, map_size, chunk_size, num_jobs
    ))

    return results



from setuptools import setup
import os

requirements = [
    "appdirs",
    "fiona",
    "numpy",
    "pyproj",
    "Rtree",
    "rasterio",
    "rasterstats",
    "shapely",
]

setup(
    version="1.0.3",
    author="Chris Mutel",
    author_email="cmutel@gmail.com",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
    package_data={'pandarus': [
        "tests/data/*.*",
    ]},
    install_requires=[] if os.environ.get('READTHEDOCS') else requirements,
    license=open('LICENSE', encoding='utf-8').read(),
    long_description=open('README.md', encoding='utf-8').read(),
    name='pandarus',
    packages=["pandarus"],
    url="https://github.com/cmutel/pandarus",
)
